@@ -61,6 +61,9 @@ static char *password = NULL;
 static int sslverify = 1;
 static int timeout = 0;
 
+/* Use '-D curl.verbose=1' to set. */
+int curl_debug_verbose = 0;
+
 static void
 curl_load (void)
 {
@@ -86,22 +89,8 @@ curl_config (const char *key, const char *value)
 
   else if (strcmp (key, "password") == 0) {
     free (password);
-    password = NULL;
-
-    if (strcmp (value, "-") == 0) {
-      ssize_t r;
-
-      printf ("password: ");
-      r = getline (&password, NULL, stdin);
-      if (r == -1) {
-        nbdkit_error ("could not read password from stdin: %m");
-        return -1;
-      }
-      if (password && r > 0 && password[r-1] == '\n')
-        password[r-1] = '\0';
-    }
-    else
-      password = strdup (value);
+    if (nbdkit_read_password (value, &password) == -1)
+      return -1;
   }
 
   else if (strcmp (key, "sslverify") == 0) {
@@ -190,6 +179,12 @@ curl_open (int readonly)
 
   nbdkit_debug ("opened libcurl easy handle");
 
+  /* Note this writes the output to stderr directly.  We should
+   * consider using CURLOPT_DEBUGFUNCTION so we can handle it with
+   * nbdkit_debug.
+   */
+  curl_easy_setopt (h->c, CURLOPT_VERBOSE, curl_debug_verbose);
+
   curl_easy_setopt (h->c, CURLOPT_ERRORBUFFER, h->errbuf);
 
   r = curl_easy_setopt (h->c, CURLOPT_URL, url);
