@@ -40,6 +40,7 @@
 
 #include <stdio.h>
 #include <stdlib.h>
+#include <stdint.h>
 #include <string.h>
 #include <fcntl.h>
 #include <unistd.h>
@@ -48,14 +49,37 @@
 
 #include <nbdkit-plugin.h>
 
+#ifndef O_CLOEXEC
+#define O_CLOEXEC 0
+#endif
+
 static char *filename = NULL;
 
+/* A debug flag which can be set on the command line using
+ * '-D example2.extra=1' to enable very verbose debugging to help
+ * developers.  Use the debug flags for extra debugging which would
+ * only be useful for the original developers of the plugin.  For
+ * ordinary debugging, just use nbdkit_debug and enable messages with
+ * the -v flag on the command line.
+ */
+int example2_debug_extra = 0;
+
 static void
 example2_unload (void)
 {
   free (filename);
 }
 
+/* If you want to display extra information about the plugin when
+ * the user does ‘nbdkit example2 --dump-plugin’ then you can print
+ * ‘key=value’ lines here.
+ */
+static void
+example2_dump_plugin (void)
+{
+  printf ("example2_extra=hello\n");
+}
+
 /* Called for each key=value passed on the command line.  This plugin
  * only accepts file=<filename>, which is required.
  */
@@ -64,7 +88,7 @@ example2_config (const char *key, const char *value)
 {
   if (strcmp (key, "file") == 0) {
     /* See FILENAMES AND PATHS in nbdkit-plugin(3). */
-    filename = nbdkit_absolute_path (value);
+    filename = nbdkit_realpath (value);
     if (!filename)
       return -1;
   }
@@ -151,6 +175,16 @@ example2_get_size (void *handle)
     return -1;
   }
 
+  /* Use the debug flags for extra debugging which would only be
+   * useful for the original developers of the plugin.  For ordinary
+   * debugging, just use nbdkit_debug and enable messages with the -v
+   * flag on the command line.  This is a contrived example of how to
+   * use debug flags.
+   */
+  if (example2_debug_extra)
+    nbdkit_debug ("extra debugging: statbuf.st_size = %jd\n",
+                  (intmax_t)statbuf.st_size);
+
   return statbuf.st_size;
 }
 
@@ -182,6 +216,7 @@ static struct nbdkit_plugin plugin = {
   .name              = "example2",
   .version           = PACKAGE_VERSION,
   .unload            = example2_unload,
+  .dump_plugin       = example2_dump_plugin,
   .config            = example2_config,
   .config_complete   = example2_config_complete,
   .config_help       = example2_config_help,
@@ -189,6 +224,10 @@ static struct nbdkit_plugin plugin = {
   .close             = example2_close,
   .get_size          = example2_get_size,
   .pread             = example2_pread,
+  /* In this plugin, errno is preserved properly along error return
+   * paths from failed system calls.
+   */
+  .errno_is_preserved = 1,
 };
 
 NBDKIT_REGISTER_PLUGIN(plugin)
