@@ -1,5 +1,5 @@
 /* nbdkit
- * Copyright (C) 2013-2014 Red Hat Inc.
+ * Copyright (C) 2013-2018 Red Hat Inc.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -169,7 +169,7 @@ perl_dump_plugin (void)
 {
   dSP;
 
-  if (callback_defined ("dump_plugin")) {
+  if (script && callback_defined ("dump_plugin")) {
     ENTER;
     SAVETMPS;
     PUSHMARK (SP);
