@@ -48,16 +48,14 @@
 #include <sys/wait.h>
 #include <errno.h>
 #include <assert.h>
-
-#ifdef HAVE_SYS_PRCTL_H
-#include <sys/prctl.h>
-#endif
+#include <syslog.h>
 
 #include <pthread.h>
 
 #include <dlfcn.h>
 
 #include "internal.h"
+#include "exit-with-parent.h"
 
 #define FIRST_SOCKET_ACTIVATION_FD 3 /* defined by systemd ABI */
 
@@ -75,10 +73,12 @@ static uid_t parseuser (const char *);
 static gid_t parsegroup (const char *);
 static unsigned int get_socket_activation (void);
 
+struct debug_flag *debug_flags; /* -D */
 int exit_with_parent;           /* --exit-with-parent */
 const char *exportname;         /* -e */
 int foreground;                 /* -f */
 const char *ipaddr;             /* -i */
+enum log_to log_to = LOG_TO_DEFAULT; /* --log */
 int newstyle = 1;               /* 0 = -o, 1 = -n */
 char *pidfile;                  /* -P */
 const char *port;               /* -p */
@@ -104,6 +104,9 @@ volatile int quit;
 int quit_fd;
 static int write_quit_fd;
 
+/* True if we forked into the background (used to control log messages). */
+int forked_into_background;
+
 /* The currently loaded plugin. */
 struct backend *backend;
 
@@ -116,6 +119,7 @@ enum {
   DUMP_PLUGIN_OPTION,
   EXIT_WITH_PARENT_OPTION,
   FILTER_OPTION,
+  LOG_OPTION,
   LONG_OPTIONS_OPTION,
   RUN_OPTION,
   SELINUX_LABEL_OPTION,
@@ -126,8 +130,9 @@ enum {
   TLS_VERIFY_PEER_OPTION,
 };
 
-static const char *short_options = "e:fg:i:nop:P:rst:u:U:vV";
+static const char *short_options = "D:e:fg:i:nop:P:rst:u:U:vV";
 static const struct option long_options[] = {
+  { "debug",            required_argument, NULL, 'D' },
   { "dump-config",      no_argument,       NULL, DUMP_CONFIG_OPTION },
   { "dump-plugin",      no_argument,       NULL, DUMP_PLUGIN_OPTION },
   { "exit-with-parent", no_argument,       NULL, EXIT_WITH_PARENT_OPTION },
@@ -141,6 +146,7 @@ static const struct option long_options[] = {
   { "help",             no_argument,       NULL, HELP_OPTION },
   { "ip-addr",          required_argument, NULL, 'i' },
   { "ipaddr",           required_argument, NULL, 'i' },
+  { "log",              required_argument, NULL, LOG_OPTION },
   { "long-options",     no_argument,       NULL, LONG_OPTIONS_OPTION },
   { "new-style",        no_argument,       NULL, 'n' },
   { "newstyle",         no_argument,       NULL, 'n' },
@@ -172,16 +178,10 @@ static void
 usage (void)
 {
   /* --{short,long}-options remain undocumented */
-  printf ("nbdkit [--dump-config] [--dump-plugin]\n"
-          "       [-e EXPORTNAME] [--exit-with-parent] [-f]\n"
-          "       [--filter=FILTER ...] [-g GROUP] [-i IPADDR]\n"
-          "       [--newstyle] [--oldstyle] [-P PIDFILE] [-p PORT] [-r]\n"
-          "       [--run CMD] [-s] [--selinux-label LABEL] [-t THREADS]\n"
-          "       [--tls=off|on|require] [--tls-certificates /path/to/certificates]\n"
-          "       [--tls-psk /path/to/pskfile] [--tls-verify-peer]\n"
-          "       [-U SOCKET] [-u USER] [-v] [-V]\n"
-          "       PLUGIN [key=value [key=value [...]]]\n"
-          "\n"
+  printf (
+#include "synopsis.c"
+  );
+  printf ("\n"
           "Please read the nbdkit(1) manual page for full usage.\n");
 }
 
@@ -252,6 +252,51 @@ main (int argc, char *argv[])
       break;
 
     switch (c) {
+    case 'D':
+      {
+        const char *p, *q;
+        struct debug_flag *flag;
+
+        /* Debug Flag must be "NAME.FLAG=N".
+         *                     ^    ^    ^
+         *                optarg    p    q  (after +1 adjustment below)
+         */
+        p = strchr (optarg, '.');
+        q = strchr (optarg, '=');
+        if (p == NULL || q == NULL) {
+        bad_debug_flag:
+          fprintf (stderr,
+                   "%s: -D (Debug Flag) must have the format NAME.FLAG=N\n",
+                   program_name);
+          exit (EXIT_FAILURE);
+        }
+        p++;                    /* +1 adjustment */
+        q++;
+
+        if (p - optarg <= 1) goto bad_debug_flag; /* NAME too short */
+        if (p > q) goto bad_debug_flag;
+        if (q - p <= 1) goto bad_debug_flag; /* FLAG too short */
+        if (*q == '\0') goto bad_debug_flag; /* N too short */
+
+        flag = malloc (sizeof *flag);
+        if (flag == NULL) {
+        debug_flag_perror:
+          perror ("malloc");
+          exit (EXIT_FAILURE);
+        }
+        flag->name = strndup (optarg, p-optarg-1);
+        if (!flag->name) goto debug_flag_perror;
+        flag->flag = strndup (p, q-p-1);
+        if (!flag->flag) goto debug_flag_perror;
+        if (sscanf (q, "%d", &flag->value) != 1) goto bad_debug_flag;
+        flag->used = 0;
+
+        /* Add flag to the linked list. */
+        flag->next = debug_flags;
+        debug_flags = flag;
+      }
+      break;
+
     case DUMP_CONFIG_OPTION:
       dump_config ();
       exit (EXIT_SUCCESS);
@@ -261,7 +306,7 @@ main (int argc, char *argv[])
       break;
 
     case EXIT_WITH_PARENT_OPTION:
-#ifdef PR_SET_PDEATHSIG
+#ifdef HAVE_EXIT_WITH_PARENT
       exit_with_parent = 1;
       foreground = 1;
       break;
@@ -286,6 +331,18 @@ main (int argc, char *argv[])
       }
       break;
 
+    case LOG_OPTION:
+      if (strcmp (optarg, "stderr") == 0)
+        log_to = LOG_TO_STDERR;
+      else if (strcmp (optarg, "syslog") == 0)
+        log_to = LOG_TO_SYSLOG;
+      else {
+        fprintf (stderr, "%s: --log must be \"stderr\" or \"syslog\"\n",
+                 program_name);
+        exit (EXIT_FAILURE);
+      }
+      break;
+
     case LONG_OPTIONS_OPTION:
       for (i = 0; long_options[i].name != NULL; ++i) {
         if (strcmp (long_options[i].name, "long-options") != 0 &&
@@ -505,6 +562,10 @@ main (int argc, char *argv[])
    */
   umask (0022);
 
+  /* If we will or might use syslog. */
+  if (log_to == LOG_TO_SYSLOG || log_to == LOG_TO_DEFAULT)
+    openlog (program_name, LOG_PID, 0);
+
   /* Initialize TLS. */
   crypto_init (tls_set_on_cli);
   assert (tls != -1);
@@ -512,10 +573,10 @@ main (int argc, char *argv[])
   /* Implement --exit-with-parent early in case plugin initialization
    * takes a long time and the parent exits during that time.
    */
-#ifdef PR_SET_PDEATHSIG
+#ifdef HAVE_EXIT_WITH_PARENT
   if (exit_with_parent) {
-    if (prctl (PR_SET_PDEATHSIG, SIGTERM) == -1) {
-      perror ("prctl: PR_SET_PDEATHSIG");
+    if (set_exit_with_parent () == -1) {
+      perror ("nbdkit: --exit-with-parent");
       exit (EXIT_FAILURE);
     }
   }
@@ -575,6 +636,23 @@ main (int argc, char *argv[])
     filter_filenames = t->next;
     free (t);
   }
+
+  /* Check all debug flags were used, and free them. */
+  while (debug_flags != NULL) {
+    struct debug_flag *next = debug_flags->next;
+
+    if (!debug_flags->used) {
+      fprintf (stderr, "%s: debug flag -D %s.%s was not used\n",
+               program_name, debug_flags->name, debug_flags->flag);
+      exit (EXIT_FAILURE);
+    }
+    free (debug_flags->name);
+    free (debug_flags->flag);
+    free (debug_flags);
+    debug_flags = next;
+  }
+
+  /* Select a thread model. */
   lock_init_thread_model ();
 
   if (help) {
@@ -988,6 +1066,7 @@ fork_into_background (void)
   if (!verbose)
     dup2 (1, 2);
 
+  forked_into_background = 1;
   debug ("forked into background (new pid = %d)", getpid ());
 }
 
