@@ -43,7 +43,6 @@
 
 #include <dlfcn.h>
 
-#include "nbdkit-plugin.h"
 #include "internal.h"
 
 /* Maximum read or write request that we will handle. */
@@ -70,11 +69,12 @@ plugin_free (struct backend *b)
    */
   lock_unload ();
 
-  debug ("%s: unload", p->filename);
+  debug ("%s: unload", p->name);
   if (p->plugin.unload)
     p->plugin.unload ();
 
-  dlclose (p->dl);
+  if (DO_DLCLOSE)
+    dlclose (p->dl);
   free (p->filename);
 
   unlock_unload ();
@@ -134,7 +134,7 @@ plugin_dump_fields (struct backend *b)
   struct backend_plugin *p = container_of (b, struct backend_plugin, backend);
   char *path;
 
-  path = nbdkit_absolute_path (p->filename);
+  path = nbdkit_realpath (p->filename);
   printf ("path=%s\n", path);
   free (path);
 
@@ -181,6 +181,12 @@ plugin_dump_fields (struct backend *b)
   HAS (can_flush);
   HAS (is_rotational);
   HAS (can_trim);
+  HAS (_pread_old);
+  HAS (_pwrite_old);
+  HAS (_flush_old);
+  HAS (_trim_old);
+  HAS (_zero_old);
+  HAS (can_fua);
   HAS (pread);
   HAS (pwrite);
   HAS (flush);
@@ -198,8 +204,7 @@ plugin_config (struct backend *b, const char *key, const char *value)
 {
   struct backend_plugin *p = container_of (b, struct backend_plugin, backend);
 
-  debug ("%s: config key=%s, value=%s",
-         p->filename, key, value);
+  debug ("%s: config key=%s, value=%s", p->name, key, value);
 
   if (p->plugin.config == NULL) {
     fprintf (stderr, "%s: %s: this plugin does not need command line configuration\n"
@@ -218,7 +223,7 @@ plugin_config_complete (struct backend *b)
 {
   struct backend_plugin *p = container_of (b, struct backend_plugin, backend);
 
-  debug ("%s: config_complete", p->filename);
+  debug ("%s: config_complete", p->name);
 
   if (!p->plugin.config_complete)
     return;
@@ -227,14 +232,6 @@ plugin_config_complete (struct backend *b)
     exit (EXIT_FAILURE);
 }
 
-static int
-plugin_errno_is_preserved (struct backend *b)
-{
-  struct backend_plugin *p = container_of (b, struct backend_plugin, backend);
-
-  return p->plugin.errno_is_preserved;
-}
-
 static int
 plugin_open (struct backend *b, struct connection *conn, int readonly)
 {
@@ -244,7 +241,7 @@ plugin_open (struct backend *b, struct connection *conn, int readonly)
   assert (connection_get_handle (conn, 0) == NULL);
   assert (p->plugin.open != NULL);
 
-  debug ("%s: open readonly=%d", p->filename, readonly);
+  debug ("%s: open readonly=%d", p->name, readonly);
 
   handle = p->plugin.open (readonly);
   if (!handle)
@@ -310,7 +307,7 @@ plugin_can_write (struct backend *b, struct connection *conn)
   if (p->plugin.can_write)
     return p->plugin.can_write (connection_get_handle (conn, 0));
   else
-    return p->plugin.pwrite != NULL;
+    return p->plugin.pwrite || p->plugin._pwrite_old;
 }
 
 static int
@@ -325,7 +322,7 @@ plugin_can_flush (struct backend *b, struct connection *conn)
   if (p->plugin.can_flush)
     return p->plugin.can_flush (connection_get_handle (conn, 0));
   else
-    return p->plugin.flush != NULL;
+    return p->plugin.flush || p->plugin._flush_old;
 }
 
 static int
@@ -355,49 +352,134 @@ plugin_can_trim (struct backend *b, struct connection *conn)
   if (p->plugin.can_trim)
     return p->plugin.can_trim (connection_get_handle (conn, 0));
   else
-    return p->plugin.trim != NULL;
+    return p->plugin.trim || p->plugin._trim_old;
 }
 
 static int
-plugin_pread (struct backend *b, struct connection *conn,
-              void *buf, uint32_t count, uint64_t offset, uint32_t flags)
+plugin_can_zero (struct backend *b, struct connection *conn)
 {
   struct backend_plugin *p = container_of (b, struct backend_plugin, backend);
 
   assert (connection_get_handle (conn, 0));
-  assert (p->plugin.pread != NULL);
+
+  debug ("can_zero");
+
+  /* Note the special case here: the plugin's .can_zero controls only
+   * whether we call .zero; while the backend expects .can_zero to
+   * return whether to advertise zero support.  Since we ALWAYS know
+   * how to fall back to .pwrite in plugin_zero(), we ignore the
+   * difference between the plugin's true or false return, and only
+   * call it to catch a -1 failure during negotiation.  */
+  if (p->plugin.can_zero &&
+      p->plugin.can_zero (connection_get_handle (conn, 0)) == -1)
+    return -1;
+  return plugin_can_write (b, conn);
+}
+
+static int
+plugin_can_fua (struct backend *b, struct connection *conn)
+{
+  struct backend_plugin *p = container_of (b, struct backend_plugin, backend);
+  int r;
+
+  assert (connection_get_handle (conn, 0));
+
+  debug ("can_fua");
+
+  /* The plugin must use API version 2 and have .can_fua return
+     NBDKIT_FUA_NATIVE before we will pass the FUA flag on. */
+  if (p->plugin.can_fua) {
+    r = p->plugin.can_fua (connection_get_handle (conn, 0));
+    if (r > NBDKIT_FUA_EMULATE && p->plugin._api_version == 1)
+      r = NBDKIT_FUA_EMULATE;
+    return r;
+  }
+  /* We intend to call .flush even if .can_flush returns false. */
+  if (p->plugin.flush || p->plugin._flush_old)
+    return NBDKIT_FUA_EMULATE;
+  return NBDKIT_FUA_NONE;
+}
+
+/* Plugins and filters can call this to set the true errno, in cases
+ * where !errno_is_preserved.
+ */
+void
+nbdkit_set_error (int err)
+{
+  threadlocal_set_error (err);
+}
+
+/* Grab the appropriate error value.
+ */
+static int
+get_error (struct backend_plugin *p)
+{
+  int ret = threadlocal_get_error ();
+
+  if (!ret && p->plugin.errno_is_preserved)
+    ret = errno;
+  return ret ? ret : EIO;
+}
+
+static int
+plugin_pread (struct backend *b, struct connection *conn,
+              void *buf, uint32_t count, uint64_t offset, uint32_t flags,
+              int *err)
+{
+  struct backend_plugin *p = container_of (b, struct backend_plugin, backend);
+  int r;
+
+  assert (connection_get_handle (conn, 0));
+  assert (p->plugin.pread || p->plugin._pread_old);
   assert (!flags);
 
   debug ("pread count=%" PRIu32 " offset=%" PRIu64, count, offset);
 
-  return p->plugin.pread (connection_get_handle (conn, 0), buf, count, offset);
+  if (p->plugin.pread)
+    r = p->plugin.pread (connection_get_handle (conn, 0), buf, count, offset,
+                         0);
+  else
+    r = p->plugin._pread_old (connection_get_handle (conn, 0), buf, count,
+                              offset);
+  if (r == -1)
+    *err = get_error (p);
+  return r;
 }
 
 static int
-plugin_flush (struct backend *b, struct connection *conn, uint32_t flags)
+plugin_flush (struct backend *b, struct connection *conn, uint32_t flags,
+              int *err)
 {
   struct backend_plugin *p = container_of (b, struct backend_plugin, backend);
+  int r;
 
   assert (connection_get_handle (conn, 0));
   assert (!flags);
 
   debug ("flush");
 
-  if (p->plugin.flush != NULL)
-    return p->plugin.flush (connection_get_handle (conn, 0));
+  if (p->plugin.flush)
+    r = p->plugin.flush (connection_get_handle (conn, 0), 0);
+  else if (p->plugin._flush_old)
+    r = p->plugin._flush_old (connection_get_handle (conn, 0));
   else {
-    errno = EINVAL;
+    *err = EINVAL;
     return -1;
   }
+  if (r == -1)
+    *err = get_error (p);
+  return r;
 }
 
 static int
 plugin_pwrite (struct backend *b, struct connection *conn,
-               const void *buf, uint32_t count, uint64_t offset, uint32_t flags)
+               const void *buf, uint32_t count, uint64_t offset, uint32_t flags,
+               int *err)
 {
   int r;
   struct backend_plugin *p = container_of (b, struct backend_plugin, backend);
   bool fua = flags & NBDKIT_FLAG_FUA;
+  bool need_flush = false;
 
   assert (connection_get_handle (conn, 0));
   assert (!(flags & ~NBDKIT_FLAG_FUA));
@@ -405,27 +487,35 @@ plugin_pwrite (struct backend *b, struct connection *conn,
   debug ("pwrite count=%" PRIu32 " offset=%" PRIu64 " fua=%d", count, offset,
          fua);
 
-  if (p->plugin.pwrite != NULL)
-    r = p->plugin.pwrite (connection_get_handle (conn, 0),
-                          buf, count, offset);
+  if (fua && plugin_can_fua (b, conn) != NBDKIT_FUA_NATIVE) {
+    flags &= ~NBDKIT_FLAG_FUA;
+    need_flush = true;
+  }
+  if (p->plugin.pwrite)
+    r = p->plugin.pwrite (connection_get_handle (conn, 0), buf, count, offset,
+                          flags);
+  else if (p->plugin._pwrite_old)
+    r = p->plugin._pwrite_old (connection_get_handle (conn, 0),
+                               buf, count, offset);
   else {
-    errno = EROFS;
+    *err = EROFS;
     return -1;
   }
-  if (r == 0 && fua) {
-    assert (p->plugin.flush);
-    r = plugin_flush (b, conn, 0);
-  }
+  if (r != -1 && need_flush)
+    r = plugin_flush (b, conn, 0, err);
+  if (r == -1)
+    *err = get_error (p);
   return r;
 }
 
 static int
 plugin_trim (struct backend *b, struct connection *conn,
-             uint32_t count, uint64_t offset, uint32_t flags)
+             uint32_t count, uint64_t offset, uint32_t flags, int *err)
 {
   int r;
   struct backend_plugin *p = container_of (b, struct backend_plugin, backend);
   bool fua = flags & NBDKIT_FLAG_FUA;
+  bool need_flush = false;
 
   assert (connection_get_handle (conn, 0));
   assert (!(flags & ~NBDKIT_FLAG_FUA));
@@ -433,30 +523,38 @@ plugin_trim (struct backend *b, struct connection *conn,
   debug ("trim count=%" PRIu32 " offset=%" PRIu64 " fua=%d", count, offset,
          fua);
 
-  if (p->plugin.trim != NULL)
-    r = p->plugin.trim (connection_get_handle (conn, 0), count, offset);
+  if (fua && plugin_can_fua (b, conn) != NBDKIT_FUA_NATIVE) {
+    flags &= ~NBDKIT_FLAG_FUA;
+    need_flush = true;
+  }
+  if (p->plugin.trim)
+    r = p->plugin.trim (connection_get_handle (conn, 0), count, offset, flags);
+  else if (p->plugin._trim_old)
+    r = p->plugin._trim_old (connection_get_handle (conn, 0), count, offset);
   else {
-    errno = EINVAL;
+    *err = EINVAL;
     return -1;
   }
-  if (r == 0 && fua) {
-    assert (p->plugin.flush);
-    r = plugin_flush (b, conn, 0);
-  }
+  if (r != -1 && need_flush)
+    r = plugin_flush (b, conn, 0, err);
+  if (r == -1)
+    *err = get_error (p);
   return r;
 }
 
 static int
 plugin_zero (struct backend *b, struct connection *conn,
-             uint32_t count, uint64_t offset, uint32_t flags)
+             uint32_t count, uint64_t offset, uint32_t flags, int *err)
 {
   struct backend_plugin *p = container_of (b, struct backend_plugin, backend);
   char *buf;
   uint32_t limit;
-  int result;
-  int err = 0;
-  int may_trim = (flags & NBDKIT_FLAG_MAY_TRIM) != 0;
+  int r = -1;
+  bool may_trim = flags & NBDKIT_FLAG_MAY_TRIM;
   bool fua = flags & NBDKIT_FLAG_FUA;
+  bool emulate = false;
+  bool need_flush = false;
+  int can_zero = 1; /* TODO cache this per-connection? */
 
   assert (connection_get_handle (conn, 0));
   assert (!(flags & ~(NBDKIT_FLAG_MAY_TRIM | NBDKIT_FLAG_FUA)));
@@ -464,50 +562,62 @@ plugin_zero (struct backend *b, struct connection *conn,
   debug ("zero count=%" PRIu32 " offset=%" PRIu64 " may_trim=%d fua=%d",
          count, offset, may_trim, fua);
 
+  if (fua && plugin_can_fua (b, conn) != NBDKIT_FUA_NATIVE) {
+    flags &= ~NBDKIT_FLAG_FUA;
+    need_flush = true;
+  }
   if (!count)
     return 0;
-  if (p->plugin.zero) {
+  if (p->plugin.can_zero) {
+    can_zero = p->plugin.can_zero (connection_get_handle (conn, 0));
+    assert (can_zero != -1);
+  }
+
+  if (can_zero) {
     errno = 0;
-    result = p->plugin.zero (connection_get_handle (conn, 0),
-                             count, offset, may_trim);
-    if (result == -1) {
-      err = threadlocal_get_error ();
-      if (!err && plugin_errno_is_preserved (b))
-        err = errno;
-    }
-    if (result == 0 || err != EOPNOTSUPP)
+    if (p->plugin.zero)
+      r = p->plugin.zero (connection_get_handle (conn, 0), count, offset,
+                          flags);
+    else if (p->plugin._zero_old)
+      r = p->plugin._zero_old (connection_get_handle (conn, 0), count, offset,
+                               may_trim);
+    else
+      emulate = true;
+    if (r == -1)
+      *err = emulate ? EOPNOTSUPP : get_error (p);
+    if (r == 0 || *err != EOPNOTSUPP)
       goto done;
   }
 
-  assert (p->plugin.pwrite);
+  assert (p->plugin.pwrite || p->plugin._pwrite_old);
+  flags &= ~NBDKIT_FLAG_MAY_TRIM;
   threadlocal_set_error (0);
   limit = count < MAX_REQUEST_SIZE ? count : MAX_REQUEST_SIZE;
   buf = calloc (limit, 1);
   if (!buf) {
-    errno = ENOMEM;
+    *err = ENOMEM;
     return -1;
   }
 
   while (count) {
-    result = p->plugin.pwrite (connection_get_handle (conn, 0),
-                               buf, limit, offset);
-    if (result < 0)
+    r = plugin_pwrite (b, conn, buf, limit, offset, flags, err);
+    if (r == -1)
       break;
     count -= limit;
     if (count < limit)
       limit = count;
   }
 
-  err = errno;
+  *err = errno;
   free (buf);
-  errno = err;
+  errno = *err;
 
  done:
-  if (!result && fua) {
-    assert (p->plugin.flush);
-    result = plugin_flush (b, conn, 0);
-  }
-  return result;
+  if (r != -1 && need_flush)
+    r = plugin_flush (b, conn, 0, err);
+  if (r == -1)
+    *err = get_error (p);
+  return r;
 }
 
 static struct backend plugin_functions = {
@@ -520,7 +630,6 @@ static struct backend plugin_functions = {
   .dump_fields = plugin_dump_fields,
   .config = plugin_config,
   .config_complete = plugin_config_complete,
-  .errno_is_preserved = plugin_errno_is_preserved,
   .open = plugin_open,
   .prepare = plugin_prepare,
   .finalize = plugin_finalize,
@@ -530,6 +639,8 @@ static struct backend plugin_functions = {
   .can_flush = plugin_can_flush,
   .is_rotational = plugin_is_rotational,
   .can_trim = plugin_can_trim,
+  .can_zero = plugin_can_zero,
+  .can_fua = plugin_can_fua,
   .pread = plugin_pread,
   .pwrite = plugin_pwrite,
   .flush = plugin_flush,
@@ -573,7 +684,7 @@ plugin_register (size_t index, const char *filename,
   }
 
   /* Check for incompatible future versions. */
-  if (plugin->_api_version != 1) {
+  if (plugin->_api_version < 0 || plugin->_api_version > 2) {
     fprintf (stderr, "%s: %s: plugin is incompatible with this version of nbdkit (_api_version = %d)\n",
              program_name, p->filename, plugin->_api_version);
     exit (EXIT_FAILURE);
@@ -608,7 +719,7 @@ plugin_register (size_t index, const char *filename,
              program_name, p->filename);
     exit (EXIT_FAILURE);
   }
-  if (p->plugin.pread == NULL) {
+  if (p->plugin.pread == NULL && p->plugin._pread_old == NULL) {
     fprintf (stderr, "%s: %s: plugin must have a .pread callback\n",
              program_name, p->filename);
     exit (EXIT_FAILURE);
@@ -641,10 +752,49 @@ plugin_register (size_t index, const char *filename,
 
   debug ("registered plugin %s (name %s)", p->filename, p->name);
 
+  /* Set debug flags before calling load. */
+  set_debug_flags (dl, p->name);
+
   /* Call the on-load callback if it exists. */
-  debug ("%s: load", p->filename);
+  debug ("%s: load", p->name);
   if (p->plugin.load)
     p->plugin.load ();
 
   return (struct backend *) p;
 }
+
+/* Set all debug flags which apply to this plugin (also used by filters). */
+void
+set_debug_flags (void *dl, const char *name)
+{
+  struct debug_flag *flag;
+
+  for (flag = debug_flags; flag != NULL; flag = flag->next) {
+    if (!flag->used && strcmp (name, flag->name) == 0) {
+      CLEANUP_FREE char *var = NULL;
+      int *sym;
+
+      /* Synthesize the name of the variable. */
+      if (asprintf (&var, "%s_debug_%s", name, flag->flag) == -1) {
+        perror ("asprintf");
+        exit (EXIT_FAILURE);
+      }
+
+      /* Find the symbol. */
+      sym = dlsym (dl, var);
+      if (sym == NULL) {
+        fprintf (stderr,
+                 "%s: -D %s.%s: %s does not contain a "
+                 "global variable called %s\n",
+                 program_name, name, flag->flag, name, var);
+        exit (EXIT_FAILURE);
+      }
+
+      /* Set the flag. */
+      *sym = flag->value;
+
+      /* Mark this flag as used. */
+      flag->used = 1;
+    }
+  }
+}
