@@ -1,4 +1,4 @@
-#!/bin/bash -
+#!/usr/bin/env bash
 # nbdkit
 # Copyright (C) 2018 Red Hat Inc.
 # All rights reserved.
@@ -32,6 +32,7 @@
 # SUCH DAMAGE.
 
 set -e
+set -x
 
 files="cow-base.img cow-diff.qcow2 cow.sock cow.pid"
 rm -f $files
@@ -41,7 +42,7 @@ guestfish -N cow-base.img=fs exit
 lastmod="$(stat -c "%y" cow-base.img)"
 
 # Run nbdkit with a COW overlay.
-nbdkit -P cow.pid -U cow.sock --filter cow file file=cow-base.img
+nbdkit -P cow.pid -U cow.sock --filter=cow file file=cow-base.img
 
 # We may have to wait a short time for the pid file to appear.
 for i in `seq 1 10`; do
@@ -61,6 +62,8 @@ pid="$(cat cow.pid)"
 cleanup ()
 {
     status=$?
+    trap '' INT QUIT TERM EXIT ERR
+    echo $0: cleanup: exit code $status
 
     kill $pid
     rm -f $files
@@ -70,7 +73,7 @@ cleanup ()
 trap cleanup INT QUIT TERM EXIT ERR
 
 # Write some data into the overlay.
-guestfish --format=raw -a 'nbd://?socket=cow.sock' -m /dev/sda1 <<EOF
+guestfish --format=raw -a "nbd://?socket=$PWD/cow.sock" -m /dev/sda1 <<EOF
   fill-dir / 10000
   fill-pattern "abcde" 5M /large
   write /hello "hello, world"
