@@ -1,4 +1,4 @@
-#!/bin/bash -
+#!/usr/bin/env bash
 # nbdkit
 # Copyright (C) 2014 Red Hat Inc.
 # All rights reserved.
@@ -64,7 +64,8 @@ for p in $plugins; do
             # VDDK won't run without special environment variables
             # being set, so ignore it.
             ;;
-        perl-valgrind | python-valgrind | ruby-valgrind | tcl-valgrind | \
+        lua-valgrind | perl-valgrind | python-valgrind | \
+        ruby-valgrind | tcl-valgrind | \
         example4-valgrind | tar-valgrind)
             # Plugins written in scripting languages can't run under valgrind.
             ;;
