@@ -1,4 +1,4 @@
-#!/bin/bash -
+#!/usr/bin/env bash
 # nbdkit
 # Copyright (C) 2014 Red Hat Inc.
 # All rights reserved.
@@ -33,7 +33,7 @@
 
 set -e
 
-output="$(../src/nbdkit --help)"
+output="$(nbdkit --help)"
 if [[ ! ( "$output" =~ dump-config ) ]]; then
     echo "$0: unexpected output from nbdkit --help"
     echo "$output"
